/*============================================================================
 * Parallel file I/O
 *============================================================================*/

/*
  This file is part of the "Finite Volume Mesh" library, intended to provide
  finite volume mesh and associated fields I/O and manipulation services.

  Copyright (C) 2007-2008  EDF

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*----------------------------------------------------------------------------
 * Standard C library headers
 *----------------------------------------------------------------------------*/

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*----------------------------------------------------------------------------
 * BFT library headers
 *----------------------------------------------------------------------------*/

#include <bft_mem.h>
#include <bft_file.h>
#include <bft_error.h>
#include <bft_printf.h>

/*----------------------------------------------------------------------------
 *  Local headers
 *----------------------------------------------------------------------------*/

#include <fvm_defs.h>
#include <fvm_config_defs.h>
#include <fvm_parall.h>

/*----------------------------------------------------------------------------
 *  Header for the current file
 *----------------------------------------------------------------------------*/

#include <fvm_file.h>

/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#if 0
} /* Fake brace to force Emacs auto-indentation back to column 0 */
#endif
#endif /* __cplusplus */

/*=============================================================================
 * Macro definitions
 *============================================================================*/

/*============================================================================
 * Type definitions
 *============================================================================*/

/* FVM file descriptor */

struct _fvm_file_t {

  char              *name;         /* File name */
  fvm_file_mode_t    mode;         /* File mode */
  fvm_file_type_t    type;         /* Type (text, binary) */
  int                semantics;    /* Preferred file positioning semantics */
  int                rank;         /* MPI rank */
  int                n_ranks;      /* MPI rank */
  _Bool              swap_endian;  /* Swap big-endian and little-endian ? */

  bft_file_t        *sh;           /* Serial file handle */

#if defined(FVM_HAVE_MPI)
  MPI_Comm           comm;         /* Associated MPI communicator */
#if defined(FVM_HAVE_MPI_IO)
  MPI_File           fh;           /* MPI file handle */
  MPI_Info           info;         /* MPI file info */
  MPI_Offset         offset;       /* MPI file offset */
#endif
#endif

};

/*============================================================================
 * Private function definitions
 *============================================================================*/

/*----------------------------------------------------------------------------
 * Open a file using standard C IO.
 *
 * parameters:
 *   f     <-- pointer to file handler
 *   mode  <-- file acces mode: read, write, or append
 *
 * returns:
 *   0 in case of success, -1 in case of failure
 *----------------------------------------------------------------------------*/

static int
_file_open(fvm_file_t       *f,
           fvm_file_mode_t   mode)
{
  bft_file_type_t _type;
  bft_file_mode_t _mode;

  int retval = 0;

  assert(f != NULL);

  if (f->sh != NULL)
    return 0;

  /* The file handler exists and the corresponding file is closed */

  f->mode = mode;

  switch (f->type) {
  case FVM_FILE_TYPE_TEXT:
    _type = BFT_FILE_TYPE_TEXT;
    break;
  default:
    assert(f->type == FVM_FILE_TYPE_BINARY);
    _type = BFT_FILE_TYPE_BINARY;
  }

  switch (f->mode) {
  case FVM_FILE_MODE_APPEND:
    _mode = BFT_FILE_MODE_APPEND;
    break;
  case FVM_FILE_MODE_WRITE:
    _mode = BFT_FILE_MODE_WRITE;
    break;
  default:
    assert(f->mode == FVM_FILE_MODE_READ);
    _mode = BFT_FILE_MODE_READ;
  }

  f->sh = bft_file_open(f->name, _mode, _type);

  if (f->sh == NULL)
    retval = -1;

  return retval;
}

/*----------------------------------------------------------------------------
 * Close a file using standard C IO.
 *
 * parameters:
 *   f <-> pointer to file handler
 *
 * returns:
 *   0 in case of success, -1 in case of failure
 *----------------------------------------------------------------------------*/

static int
_file_close(fvm_file_t  *f)
{
  int retval = 0;

  f->sh = bft_file_free(f->sh);

  if (f->sh != NULL)
    retval = -1;

  return retval;
}

/*----------------------------------------------------------------------------
 * Read data to a buffer, distributing a contiguous part of it to each
 * process associated with a file.
 *
 * Each process should receive a slice of the data, and we should have:
 *   global_num_start at rank 0 = 1
 *   global_num_start at rank i+1 = global_num_end at rank i.
 * Otherwise, behavior (especially positioning for future reads) is undefined.
 *
 * This version does not use MPI-IO
 *
 * parameters:
 *   buf              --> pointer to location receiving data.
 *   size             <-- size of each item of data in bytes.
 *   global_num_start <-- global number of first slice item (1 to n numbering)
 *   global_num_end   <-- global number of past-the end slice item
 *                        (1 to n numbering)
 *   f                <-- fvm_file_t descriptor.
 *
 * returns:
 *   the (local) number of items (not bytes) sucessfully read;
 *----------------------------------------------------------------------------*/

static size_t
_file_read_slice(void        *buf,
                 size_t       size,
                 fvm_gnum_t   global_num_start,
                 fvm_gnum_t   global_num_end,
                 fvm_file_t  *f)
{
  size_t retval = 0;

  if (f->rank == 0)
    retval = bft_file_read(buf,
                           size,
                           (size_t)(global_num_end - global_num_start),
                           f->sh);

#if defined(FVM_HAVE_MPI)

  if (f->comm != MPI_COMM_NULL) {

    MPI_Status status;
    int rank_id;
    fvm_gnum_t slice_size;
    fvm_gnum_t range[2];

    /* Rank 0 reads data for other ranks from file and distributes it */

    if (f->rank == 0) {

      int dist_rank;
      int _retval;
      fvm_gnum_t _buf_size = global_num_end - global_num_start;
      unsigned char *_buf = NULL;

      for (dist_rank = 1; dist_rank < f->n_ranks; dist_rank++) {

        /* Forced synchronization */

        rank_id = dist_rank;
        MPI_Send(&rank_id, 1, MPI_INT, dist_rank, FVM_MPI_TAG, f->comm);

        /* Receive extents and resize buffer if necessary */

        MPI_Recv(range, 2, FVM_MPI_GNUM, dist_rank, FVM_MPI_TAG, f->comm,
                 &status);
        slice_size = (range[1] - range[0]);

        if (_buf == NULL || slice_size > _buf_size) {
          _buf_size = FVM_MAX(_buf_size, slice_size*2);
          BFT_REALLOC(_buf, _buf_size*size, unsigned char);
        }

        /* Read data from file */

        _retval = (int)bft_file_read(_buf, size, (size_t)slice_size, f->sh);

        /* Send to corresponding rank */

        MPI_Send(_buf, _retval*size, MPI_BYTE, dist_rank,
                 FVM_MPI_TAG, f->comm);

      } /* End of loop on distant ranks */

      BFT_FREE(_buf);

    }

    /* Other ranks receive data from rank 0 */

    else {

      int count;

      /* Forced synchronization */

      MPI_Recv(&rank_id, 1, MPI_INT, 0, FVM_MPI_TAG, f->comm, &status);

      /* Send extents */

      range[0] = global_num_start;
      range[1] = global_num_end;

      MPI_Send(range, 2, FVM_MPI_GNUM, 0, FVM_MPI_TAG, f->comm);

      /* Receive data */

      MPI_Recv(buf, (int)((range[1]-range[0])*size), MPI_BYTE, 0,
               FVM_MPI_TAG, f->comm, &status);

      MPI_Get_count(&status, MPI_BYTE, &count);
      retval = count / size;

    }

  }

#endif /* defined(FVM_HAVE_MPI) */

  return retval;
}

#if defined(FVM_HAVE_MPI_IO)

/*----------------------------------------------------------------------------
 * Output MPI error message.
 *
 * This supposes that the default MPI errorhandler is not used
 *
 * parameters:
 *   file_name  <-- file name
 *   error_code <-- associated MPI error code
 *
 * returns:
 *   0 in case of success, system error code in case of failure
 *  Fonction qui imprime un message d'erreur en cas de problme de
 *  communication MPI
 *----------------------------------------------------------------------------*/

static void
_mpi_io_error_message
(
 const char  *file_name,
 int          error_code
)
{
  char buffer[MPI_MAX_ERROR_STRING];
  int  buffer_len;

  MPI_Error_string(error_code, buffer, &buffer_len);

  bft_error(__FILE__, __LINE__, 0,
            _("MPI IO error for file: %s\n"
              "Error type: %s"), file_name, buffer);
}

/*----------------------------------------------------------------------------
 * Open a file using MPI IO.
 *
 * parameters:
 *   f     <-- pointer to file handler
 *   mode  <-- file acces mode: read, write, or append
 *
 * returns:
 *   MPI_SUCCESS in case of success, MPI error code in case of failure
 *----------------------------------------------------------------------------*/

static int
_mpi_file_open(fvm_file_t       *f,
               fvm_file_mode_t   mode)
{
  int amode = MPI_MODE_RDWR;
  MPI_Info  info = MPI_INFO_NULL;
  int retval = 0;

  assert(f != NULL);

  if (f->fh != MPI_FILE_NULL)
    return 0;

  if (f->type == FVM_FILE_TYPE_TEXT)
    bft_error(__FILE__, __LINE__, 0,
              _("MPI IO not available for text file\n(%s)."),
              f->name);

  /* Set access mode */

  f->mode = mode;

  if (f->mode == FVM_FILE_MODE_APPEND)
    amode = MPI_MODE_WRONLY | MPI_MODE_APPEND;

  else if (f->mode == FVM_FILE_MODE_WRITE)
    amode = MPI_MODE_WRONLY;

  else if (f->mode == FVM_FILE_MODE_READ)
    amode = MPI_MODE_RDONLY;

  /* Open file */

  retval = MPI_File_open(f->comm, f->name, amode, info, &(f->fh));

  if (retval != MPI_SUCCESS)
    _mpi_io_error_message(f->name, retval);

  return retval;
}

/*----------------------------------------------------------------------------
 * Close a file using MPI IO.
 *
 * parameters:
 *   f <-> pointer to file handler
 *
 * returns:
 *   MPI_SUCCESS in case of success, MPI error code in case of failure
 *----------------------------------------------------------------------------*/

static int
_mpi_file_close(fvm_file_t  *f)
{
  int retval = 0;

  assert(f != NULL);

  if (f->fh == MPI_FILE_NULL)
    return 0;

  /* Close file */

  retval = MPI_File_close(&(f->fh));

  if (retval != MPI_SUCCESS)
    _mpi_io_error_message(f->name, retval);

  return retval;
}

/*----------------------------------------------------------------------------
 * Read data to a buffer, distributing a contiguous part of it to each
 * process associated with a file.
 *
 * Each process should receive a slice of the data, and we should have:
 *   global_num_start at rank 0 = 1
 *   global_num_start at rank i+1 = global_num_end at rank i.
 * Otherwise, behavior (especially positioning for future reads) is undefined.
 *
 * There are 3 variants, depending on the semantics:
 *   _mpi_file_read_slice_eo (using explicit offsets)
 *   _mpi_file_read_slice_ip (using individual pointers, setting a file view)
 *   _mpi_file_read_slice_sp (using a shared pointer and ordered semantics)
 *
 * parameters:
 *   buf              --> pointer to location receiving data.
 *   size             <-- size of each item of data in bytes.
 *   global_num_start <-- global number of first slice item (1 to n numbering)
 *   global_num_end   <-- global number of past-the end slice item
 *                        (1 to n numbering)
 *   f                <-- fvm_file_t descriptor.
 *
 * returns:
 *   the (local) number of items (not bytes) sucessfully read;
 *----------------------------------------------------------------------------*/

static size_t
_mpi_file_read_slice_eo(void        *buf,
                        size_t       size,
                        fvm_gnum_t   global_num_start,
                        fvm_gnum_t   global_num_end,
                        fvm_file_t  *f)
{
  fvm_gnum_t global_num_end_last = global_num_end;
  size_t retval = 0;

  MPI_Offset disp;
  MPI_Status status;
  int errcode, count;

  disp = f->offset + ((global_num_start - 1) * size);
  count = (global_num_end - global_num_start)*size;

  errcode = MPI_File_read_at_all(f->fh, disp, buf, count, MPI_BYTE, &status);

  if (errcode != MPI_SUCCESS)
    _mpi_io_error_message(f->name, errcode);

  MPI_Get_count(&status, MPI_BYTE, &count);
  retval = count / size;

  MPI_Bcast(&global_num_end_last, 1, FVM_MPI_GNUM, f->n_ranks-1, f->comm);
  f->offset += ((global_num_end_last - 1) * size);

  return retval;
}

static size_t
_mpi_file_read_slice_ip(void        *buf,
                        size_t       size,
                        fvm_gnum_t   global_num_start,
                        fvm_gnum_t   global_num_end,
                        fvm_file_t  *f)
{
  char datarep[] = "native";
  fvm_gnum_t global_num_end_last = global_num_end;
  size_t retval = 0;

  MPI_Status status;
  MPI_Datatype file_type;
  int errcode, count;

  MPI_Aint lengths[1] = {(global_num_end - global_num_start) * size};
  MPI_Aint disps[1] = {(global_num_start - 1) * size};

  MPI_Type_hindexed(1, lengths, disps, MPI_BYTE, &file_type);
  MPI_Type_commit(&file_type);

  MPI_File_set_view(f->fh, f->offset, MPI_BYTE, file_type, datarep, f->info);

  errcode = MPI_File_read_all(f->fh, buf, (int)(lengths[0]), MPI_BYTE,
                              &status);

  if (errcode != MPI_SUCCESS)
    _mpi_io_error_message(f->name, errcode);

  MPI_Type_free(&file_type);

  MPI_Get_count(&status, MPI_BYTE, &count);
  retval = count / size;

  MPI_Bcast(&global_num_end_last, 1, FVM_MPI_GNUM, f->n_ranks-1, f->comm);
  f->offset += ((global_num_end_last - 1) * size);

  return retval;
}

static size_t
_mpi_file_read_slice_sp(void        *buf,
                        size_t       size,
                        fvm_gnum_t   global_num_start,
                        fvm_gnum_t   global_num_end,
                        fvm_file_t  *f)
{
  fvm_gnum_t global_num_end_last = global_num_end;
  size_t retval = 0;

  MPI_Offset disp;
  MPI_Status status;
  int errcode, count;

  disp = f->offset + ((global_num_start - 1) * size);
  count = (global_num_end - global_num_start)*size;

  errcode = MPI_File_read_ordered(f->fh, buf, count, MPI_BYTE, &status);

  if (errcode != MPI_SUCCESS)
    _mpi_io_error_message(f->name, errcode);

  MPI_Get_count(&status, MPI_BYTE, &count);
  retval = count / size;

  MPI_Bcast(&global_num_end_last, 1, FVM_MPI_GNUM, f->n_ranks-1, f->comm);
  f->offset += ((global_num_end_last - 1) * size);

  return retval;
}

#endif /* defined(FVM_HAVE_MPI_IO) */

/*=============================================================================
 * Public function definitions
 *============================================================================*/

/*----------------------------------------------------------------------------
 * Create a file descriptor and open the associated file.
 *
 * By default, data is written or read as native data. This behavior may be
 * modified by fvm_file_set_swap_endian().
 *
 * parameters:
 *   name  <-- file name
 *   mode  <-- file acces mode: read, write, or append
 *   type  <-- file type: text or binary
 *   hints <-- file I/O hints (for MPI and MPI I/O behavior)
 *
 * returns:
 *   pointer to fvm_file_t file descriptor (NULL in case of failure).
 *----------------------------------------------------------------------------*/

fvm_file_t *
fvm_file_open(const char         *name,
              fvm_file_mode_t     mode,
              fvm_file_type_t     type,
              fvm_file_hints_t    hints)
{
  int errcode = 0;
  fvm_file_t * f = NULL;

  BFT_MALLOC(f, 1, fvm_file_t);

  f->sh = NULL;

#if defined(FVM_HAVE_MPI)
  f->comm = MPI_COMM_NULL;
#if defined(FVM_HAVE_MPI_IO)
  f->fh = MPI_FILE_NULL;
  f->info = MPI_INFO_NULL;
  f->offset = 0;
#endif
#endif

  BFT_MALLOC(f->name, strlen(name) + 1, char);
  strcpy(f->name, name);

  f->mode = mode;
  f->type = type;
  f->semantics = FVM_FILE_NO_MPI_IO;
  f->rank = 0;
  f->n_ranks = 1;

  f->swap_endian = false; /* Use native endianness by default */

  /* Set communicator */

#if defined(FVM_HAVE_MPI)
  {
    MPI_Comm comm = fvm_parall_get_mpi_comm();
    if (comm != MPI_COMM_NULL) {
      MPI_Comm_size(comm, &(f->n_ranks));
      if (f->n_ranks > 1) {
        MPI_Comm_dup(comm, &(f->comm));
        MPI_Comm_rank(f->comm, &(f->rank));
      }
      else
        f->comm = MPI_COMM_NULL;
    }
  }
#endif /* defined(FVM_HAVE_MPI) */

  /* Use MPI IO ? */

#if defined(FVM_HAVE_MPI_IO)
  if (   f->type != FVM_FILE_TYPE_TEXT
      && f->comm != MPI_COMM_NULL
      && !(hints & FVM_FILE_NO_MPI_IO)) {
    int positioning_mask = (  FVM_FILE_EXPLICIT_OFFSETS
                            | FVM_FILE_INDIVIDUAL_POINTERS
                            | FVM_FILE_SHARED_POINTER);
    if (hints & positioning_mask)
      f->semantics = hints & positioning_mask;
    else
      f->semantics = FVM_FILE_INDIVIDUAL_POINTERS;
    f->semantics = f->semantics | (hints & FVM_FILE_NO_PREDISTRIBUTE);
  }
#endif

  /* Open file. In case of failure, destroy the allocated structure;
     this is only useful with a non-default error handler,
     as the program is terminated by default */

  if ((f->semantics & FVM_FILE_NO_MPI_IO) && f->rank == 0)
    errcode = _file_open(f, f->mode);

#if defined(FVM_HAVE_MPI_IO)
  else if (!(f->semantics & FVM_FILE_NO_MPI_IO))
    errcode = _mpi_file_open(f, f->mode);
#endif

  if (errcode != 0)
    f = fvm_file_free(f);

  return f;
}

/*----------------------------------------------------------------------------
 * Destroy a file descriptor and close the associated file.
 *
 * parameters:
 *   f <-> file descriptor to destroy
 *
 * returns:
 *   NULL pointer
 *----------------------------------------------------------------------------*/

fvm_file_t *
fvm_file_free(fvm_file_t  *f)
{
  fvm_file_t  *_f = f;
  int errcode = 0;

  if (_f->sh != NULL)
    errcode = _file_close(_f);

#if defined(FVM_HAVE_MPI_IO)
  else if (_f->fh != MPI_FILE_NULL)
    errcode = _mpi_file_close(_f);
#endif

#if defined(FVM_HAVE_MPI)
  if (_f->comm != MPI_COMM_NULL)
    MPI_Comm_free(&(_f->comm));
#endif

  BFT_FREE(_f->name);
  BFT_FREE(_f);

  return NULL;
}

/*----------------------------------------------------------------------------
 * Ensure that data is read or written in big-endian
 * (network standard) format.
 *
 * parameters:
 *   f <-> fvm_file_t descriptor.
 *----------------------------------------------------------------------------*/

void
fvm_file_set_big_endian(fvm_file_t  *f)
{
  unsigned  int_endian;

  /* Check if system is "big-endian" or "little-endian" */

  int_endian = 0;
  *((char *)(&int_endian)) = '\1';

  if (int_endian == 1)
    f->swap_endian = 1;

#if defined(DEBUG) && !defined(NDEBUG)

  else {
    int_endian = 0;
    *((char *) (&int_endian) + sizeof(unsigned) - 1) = '\1';
    assert(int_endian == 1);
  }

#endif
}

/*----------------------------------------------------------------------------
 * Return a file's byte-swapping behavior.
 *
 * parameters:
 *   f <-- fvm_file_t descriptor.
 *
 * returns:
 *   0 if file's endianness is the same as the system's, 1 otherwise.
 *----------------------------------------------------------------------------*/

int
fvm_file_get_swap_endian(const fvm_file_t  *f)
{
  assert(f != NULL);

  return f->swap_endian;
}

/*----------------------------------------------------------------------------
 * Set a file's byte-swapping behavior.
 *
 * Using this function assumes one is familiar with a file's coding
 * or structure; use with caution.
 *
 * parameters:
 *   f    <-> fvm_file_t descriptor.
 *   swap  --> 1 if bytes must be swapped, 0 otherwise.
 *----------------------------------------------------------------------------*/

void
fvm_file_set_swap_endian(fvm_file_t  *f,
                         int          swap)
{
  assert(f != NULL);

  f->swap_endian = swap;
}

/*----------------------------------------------------------------------------
 * Read data to a buffer, distributing it to all processes associated
 * with a file.
 *
 * parameters:
 *   buf  --> pointer to location receiving data.
 *   size <-- size of each item of data in bytes.
 *   ni   <-- number of items to read.
 *   f    <-- fvm_file_t descriptor.
 *
 * returns:
 *   the number of items (not bytes) sucessfully read;
 *----------------------------------------------------------------------------*/

size_t
fvm_file_read_global(void        *buf,
                     size_t       size,
                     size_t       ni,
                     fvm_file_t  *f)
{
  size_t retval;

  if ((f->semantics & FVM_FILE_NO_MPI_IO) && f->sh != NULL) {
    retval = bft_file_read(buf,
                           size,
                           ni,
                           f->sh);
  }

#if defined(FVM_HAVE_MPI)
  {
    if ((f->semantics & FVM_FILE_NO_MPI_IO) && f->comm != MPI_COMM_NULL)
      MPI_Bcast(buf, size*ni, MPI_BYTE, 0, f->comm);

#if defined(FVM_HAVE_MPI_IO)

    else if (!(f->semantics & FVM_FILE_NO_MPI_IO)) {

      MPI_Status status;
      int errcode, count;

      if (f->semantics & FVM_FILE_EXPLICIT_OFFSETS) {
        errcode = MPI_File_read_at_all(f->fh,
                                       f->offset,
                                       buf,
                                       size*ni,
                                       MPI_BYTE,
                                       &status);
        MPI_Get_count(&status, MPI_BYTE, &count);
      }

      else if (f->semantics & FVM_FILE_INDIVIDUAL_POINTERS) {
        errcode = MPI_File_read_all(f->fh,
                                    buf,
                                    size*ni,
                                    MPI_BYTE,
                                    &status);
        MPI_Get_count(&status, MPI_BYTE, &count);
      }

      else { /* if (f->semantics & FVM_FILE_SHARED_POINTER) */
        int aux[2];
        if (f->rank == 0) {
          aux[0] = MPI_File_read_shared(f->fh,
                                        buf,
                                        size*ni,
                                        MPI_BYTE,
                                        &status);
          MPI_Get_count(&status, MPI_BYTE, &aux[1]);
        }
        MPI_Bcast(buf, size*ni, MPI_BYTE, 0, f->comm);
        MPI_Bcast(aux, 2, MPI_INT, 0, f->comm);
        errcode = aux[0];
        count = aux[1];
      }

      if (errcode != MPI_SUCCESS)
        _mpi_io_error_message(f->name, errcode);

      retval = count / size;

      if (!(f->semantics & FVM_FILE_SHARED_POINTER))
        f->offset += count;
    }

#endif /* defined(FVM_HAVE_MPI_IO) */
  }
#endif /* defined(FVM_HAVE_MPI) */

  if (f->swap_endian == true && size > 1)
    bft_file_swap_endian(buf, buf, size, retval);

  return retval;
}

/*----------------------------------------------------------------------------
 * Read data to a buffer, distributing a contiguous part of it to each
 * process associated with a file.
 *
 * Each process should receive a slice of the data, and we should have:
 *   global_num_start at rank 0 = 1
 *   global_num_start at rank i+1 = global_num_end at rank i.
 * Otherwise, behavior (especially positioning for future reads) is undefined.
 *
 * parameters:
 *   buf              --> pointer to location receiving data.
 *   size             <-- size of each item of data in bytes.
 *   global_num_start <-- global number of first slice item (1 to n numbering)
 *   global_num_end   <-- global number of past-the end slice item
 *                        (1 to n numbering)
 *   f                <-- fvm_file_t descriptor.
 *
 * returns:
 *   the (local) number of items (not bytes) sucessfully read;
 *----------------------------------------------------------------------------*/

size_t
fvm_file_read_slice(void        *buf,
                    size_t       size,
                    fvm_gnum_t   global_num_start,
                    fvm_gnum_t   global_num_end,
                    fvm_file_t  *f)
{
  size_t retval;

  if (f->semantics & FVM_FILE_NO_MPI_IO)
    retval = _file_read_slice(buf,
                              size,
                              global_num_start,
                              global_num_end,
                              f);

#if defined(FVM_HAVE_MPI_IO)

  else if (!(f->semantics & FVM_FILE_NO_MPI_IO)) {

    if (f->semantics & FVM_FILE_EXPLICIT_OFFSETS)
      retval = _mpi_file_read_slice_eo(buf,
                                       size,
                                       global_num_start,
                                       global_num_end,
                                       f);

    else if (f->semantics & FVM_FILE_INDIVIDUAL_POINTERS)
      retval = _mpi_file_read_slice_ip(buf,
                                       size,
                                       global_num_start,
                                       global_num_end,
                                       f);

    else /* if (f->semantics & FVM_FILE_SHARED_POINTER) */
      retval = _mpi_file_read_slice_sp(buf,
                                       size,
                                       global_num_start,
                                       global_num_end,
                                       f);
  }

#endif /* defined(FVM_HAVE_MPI_IO) */

  if (f->swap_endian == true && size > 1)
    bft_file_swap_endian(buf, buf, size, retval);

  return retval;
}

/*----------------------------------------------------------------------------
 * Dump the metadata of a file structure in human readable form
 *
 * parameters:
 *   this_file <-- pointer to file
 *----------------------------------------------------------------------------*/

void
fvm_file_dump(const fvm_file_t  *this_file)
{
  const char *type_name[] = {"FVM_FILE_TYPE_TEXT",
                             "FVM_FILE_TYPE_BINARY"};
  const char *mode_name[] = {"FVM_FILE_MODE_READ",
                             "FVM_FILE_MODE_WRITE",
                             "FVM_FILE_MODE_APPEND"};

  const fvm_file_t  *f = this_file;

  if (f == NULL) {
    bft_printf(_("\n"
                 "Null file dump:\n"));
    return;
  }

  bft_printf(_("\n"
               "File name:                \"%s\"\n"
               "Access mode:              %s\n"
               "File type:                %s\n"
               "Semantics:\n"
               "  no_mpi_io:              %d\n"
               "  no_predistribute:       %d\n"
               "  explicit_offsets:       %d\n"
               "  individual_pointers:    %d\n"
               "  shared_pointer:         %d\n"
               "Rank:                     %d\n"
               "N ranks:                  %d\n"
               "Swap endian:              %d\n"
               "Serial handle:            %p\n"),
             f->name, mode_name[f->mode], type_name[f->type],
             (f->semantics & FVM_FILE_NO_MPI_IO),
             (f->semantics & FVM_FILE_NO_PREDISTRIBUTE) >> 1,
             (f->semantics & FVM_FILE_EXPLICIT_OFFSETS) >> 2,
             (f->semantics & FVM_FILE_INDIVIDUAL_POINTERS) >> 3,
             (f->semantics & FVM_FILE_SHARED_POINTER) >> 4,
             f->rank, f->n_ranks, (int)(f->swap_endian),
             f->sh);

#if defined(FVM_HAVE_MPI)
  bft_printf(_("Associated communicator:  %ul\n"),
             (unsigned long)(f->comm));
#if defined(FVM_HAVE_MPI_IO)
  bft_printf(_("MPI file handle:          %ul\n"
               "MPI file offset:          %ul\n"),
             (unsigned long)(f->fh),
             (unsigned long)(f->offset));
#endif
#endif

  bft_printf("\n");
}

/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
}
#endif /* __cplusplus */
