/*============================================================================
 * Unit test for fvm_file.c;
 *============================================================================*/

/*
  This file is part of the "Finite Volume Mesh" library, intended to provide
  finite volume mesh and associated fields I/O and manipulation services.

  Copyright (C) 2007-2008  EDF

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <bft_file.h>
#include <bft_error.h>
#include <bft_mem.h>
#include <bft_printf.h>

#include "fvm_file.h"
#include "fvm_parall.h"

/*---------------------------------------------------------------------------*/

static void
_create_test_data(void)
{
  int i;
  bft_file_t *f;

  char header[80];
  int iarray[30];
  double farray[30];

  sprintf(header, "fvm test file");
  for (i = strlen(header); i < 80; i++)
    header[i] = '\0';

  for (i = 0; i < 30; i++)
    iarray[i] = i+1;
  for (i = 0; i < 30; i++)
    farray[i] = i+1;

  f= bft_file_open("file_test_data",
                   BFT_FILE_MODE_WRITE,
                   BFT_FILE_TYPE_BINARY);

  bft_file_set_big_endian(f);

  bft_file_write(header, 1, 80, f);
  bft_file_write(iarray, sizeof(int), 30, f);
  bft_file_write(farray, sizeof(double), 30, f);

  f = bft_file_free(f);
}

/*---------------------------------------------------------------------------*/

int
main (int argc, char *argv[])
{
  char mem_trace_name[32];
  char buf[80];
  int ibuf[30];
  double dbuf[30];
  fvm_gnum_t i;
  int size = 1;
  int rank = 0;
  fvm_gnum_t slice_start, slice_end;
  fvm_file_t *f = NULL;

#if defined(FVM_HAVE_MPI)

  MPI_Status status;
  int sync = 1;

  /* Initialization */

  MPI_Init(&argc, &argv);

  fvm_parall_set_mpi_comm(MPI_COMM_WORLD);

  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  slice_start = rank * (30./size) + 1;
  slice_end   = (rank + 1) * (30./size) + 1;
  if (rank == size - 1)
    slice_end = 31;

#endif /* (FVM_HAVE_MPI) */

  if (size > 1)
    sprintf(mem_trace_name, "fvm_file_test_mem.%d", rank);
  else
    strcpy(mem_trace_name, "fvm_file_test_mem");
  bft_mem_init(mem_trace_name);

  if (rank == 0)
    _create_test_data();

  f = fvm_file_open("file_test_data",
                    FVM_FILE_MODE_READ,
                    BFT_FILE_TYPE_BINARY,
                    FVM_FILE_EXPLICIT_OFFSETS);

  fvm_file_set_big_endian(f);

  fvm_file_dump(f);

  fvm_file_read_global(buf, 1, 80, f);

  bft_printf("rank %d, buf = %s\n", rank, buf);

  for (i = 0; i < 30; i++)
    ibuf[i] = 0;

  fvm_file_read_slice(ibuf, sizeof(int), slice_start, slice_end, f);

#if defined(FVM_HAVE_MPI) /* Serialize dump */
  if (rank > 0)
    MPI_Recv(&sync, 1, MPI_INT, rank - 1, 0, MPI_COMM_WORLD, &status);
#endif

  bft_printf("\nRead by rank %d:\n\n", rank);
  for (i = slice_start; i < slice_end; i++)
    bft_printf("  ival[%d] = %d\n", (int)i, (int)(ibuf[i-slice_start]));

#if defined(FVM_HAVE_MPI)
  if (rank < size - 1)
    MPI_Send(&sync, 1, MPI_INT, rank + 1, 0, MPI_COMM_WORLD);
#endif

  fvm_file_read_slice(dbuf, sizeof(double), slice_start, slice_end, f);

#if defined(FVM_HAVE_MPI) /* Serialize dump */
  if (rank > 0)
    MPI_Recv(&sync, 1, MPI_INT, rank - 1, 0, MPI_COMM_WORLD, &status);
#endif

  bft_printf("\nRead by rank %d:\n\n", rank);
  for (i = slice_start; i < slice_end; i++)
    bft_printf("  dval[%d] = %f\n", (int)i, (dbuf[i-slice_start]));

#if defined(FVM_HAVE_MPI)
  if (rank < size - 1)
    MPI_Send(&sync, 1, MPI_INT, rank + 1, 0, MPI_COMM_WORLD);
#endif

#if defined(FVM_HAVE_MPI)
  MPI_Barrier(MPI_COMM_WORLD);
#endif
  bft_printf("barrier passed by rank %d\n", rank);

  f = fvm_file_free(f);

  /* We are finished */

  bft_mem_end();

#if defined(FVM_HAVE_MPI)

  MPI_Finalize();

#endif

  exit (EXIT_SUCCESS);
}
